 #include <errno.h>
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <libusb.h>
 #include <unistd.h>
 #include <time.h>

 #define VENDOR_ID 0x09cb
 #define PRODUCT_ID 0x1996
 
 
 static struct libusb_device_handle *devh = NULL;
 
 static int find_lvr_flirusb(void)
 {
 	devh = libusb_open_device_with_vid_pid(NULL, VENDOR_ID, PRODUCT_ID);
 	return devh ? 0 : -EIO;
 }
 
 void print_bulk_result(char ep[],char EP_error[], int r, int actual_length, unsigned char buf[])
 {
         time_t now1;
         now1 = time(NULL);
         if (r < 0) {
//		printf( "errno= %d\n", errno ); 
//		perror( "libusb_submit_transfer" ); 
                if (strcmp (EP_error, libusb_error_name(r))!=0)
                {       
                    strcpy(EP_error, libusb_error_name(r));
                    fprintf(stderr, "\n: %s >>>>>>>>>>>>>>>>>bulk transfer (in) %s: %s\n", ctime(&now1), ep , libusb_error_name(r));
                    sleep(1);
                }
                //return 1;
        } else
        {
            printf("\n: %s bulk read EP %s, actual length %d\nHEX:\n",ctime(&now1), ep ,actual_length);
            for (int i = 0; i < actual_length; i++) {
                    printf(" %02x", buf[i]);
            }
            printf("\nSTRING:\n");	
            for (int i = 0; i < actual_length; i++) {
                    if(buf[i]>31) {printf("%c", buf[i]);}
            }
            printf("\n");	
            
        } 
 }       
 
 int main(void)
 {
 	int i,r = 1;
 	r = libusb_init(NULL);
 	if (r < 0) {
 		fprintf(stderr, "failed to initialise libusb\n");
 		exit(1);
 	}
 	
  	r = find_lvr_flirusb();
 	if (r < 0) {
 		fprintf(stderr, "Could not find/open device\n");
 		goto out;
 	}
 	printf("Successfully find the Flir One G2 device\n");


// 	libusb_reset_device(devh);
    unsigned char tbuf[255]; /* Some devices choke on size > 255 */
 	
 	//SETUP: 76.0 
    //DATA0: 80 | 06 | 00 01 | 00 00 | 12 00 | e0 f4 
    memset(&tbuf[0], 0, sizeof(tbuf));

 	r= libusb_control_transfer(devh, 0x80, 0x06, 0x0100, 0x0000, tbuf, 0x12, 200);
	if (r < 0) {
 		fprintf(stderr, "failed with error %d %s\n", r, libusb_error_name(r));
 		goto out;
 	} else {
     	printf("\nlibusb_control_transfer(devh, DATA0: 80 | 06 | 00 01 | 00 00 | 12 00 \nDATA1: ");	
        for (int i = 0; i < tbuf[0]; i++) {
                    printf(" %02x", tbuf[i]);
        }
        printf("\n");
    }
 	
 	// next two transfer are not usb conform! Flir knocking? 	
 	// SETUP: 76.0 
    // DATA0: 80 06 00 02 00 00 09 00 | ae 04
    
    // the usb protocol sniffer reads 64 Bytes
    // if you set the to buffer[255] then you get this additional  bytes
    // ... 09 04 02 01 00 ff f0 01 07 09 04 02 00 02 ff f0 01 07 07 05 85 02 00 02 00 07 05 06 02 00 02 01
    memset(&tbuf[0], 0, sizeof(tbuf));
 	r= libusb_control_transfer(devh, 0x80, 0x06, 0x0200, 0x0000, tbuf, 0x09, 200);
	if (r < 0) {
 		fprintf(stderr, "failed with error %d %s\n", r, libusb_error_name(r));
 		goto out;
 	} else {
     	printf("\nlibusb_control_transfer(devh, DATA0: 80 06 00 02 00 00 09 00 \nDATA1: ");	
        for (int i = 0; i < 0x09; i++) {
                    printf(" %02x", tbuf[i]);
        }
        printf("\n");
    }
    
   
    // SETUP: 76.0 
    // DATA0: 80 06 00 02 00 00 60 00 | 80 54 
    
    memset(&tbuf[0], 0, sizeof(tbuf));
 	r= libusb_control_transfer(devh, 0x80, 0x06, 0x0200, 0x0000, tbuf, 0x60, 200);
	if (r < 0) {
 		fprintf(stderr, "failed with error %d %s\n", r, libusb_error_name(r));
 		goto out;
 	} else {
     	printf("\nlibusb_control_transfer(devh, DATA0: 80 06 00 02 00 00 60 00 \nDATA1: ");	
        for (int i = 0; i < 0x60; i++) {
                    printf(" %02x", tbuf[i]);
        }
        printf("\n");
    }
    

    // get String Descriptor (0,0),(30,01) ... (30,07) in unsorted order 2,1,3,4,5,6,7
    uint16_t langid;
    
    // get string_descriptor langid
    // 0 bmRequestType = 80
    // 1 bRequest = 06                          GET_DESCRIPTOR
    // 2 wValue 0300 type (H) index (L)          type: String Descriptor (0x03)  // http://www.beyondlogic.org/usbnutshell/usb5.shtml#StringDescriptors
    // 4 wIndex 00
    // 5 wLength 00
    
    r = libusb_get_string_descriptor(devh, 0, 0, tbuf, 0xFF);
	if (r < 4)
	{
		fprintf(stderr, "\nlibusb_get_string_descriptor 0,0 error %d %s\n", r, libusb_error_name(r));
 		 goto out;
    }
	langid = tbuf[2] | (tbuf[3] << 8);
	printf("\nlibusb_get_string_descriptor langid=%#04x\n",langid);

    unsigned char string_buffer[255];
    // see order of questions in F6c.txt
    int  a[8] = { 2,1,3,4,5,6,7};
    
	// Get all string descriptors.
	for (i = 0; i <= 6; i++)
	{
	
	    r = libusb_get_string_descriptor(devh, a[i], langid, tbuf, 0xFF);
	
		if (r >= 0)
		{
			//printf("String descriptor %u: \"%s\"\n", a[i], string_buffer);
			int si, di;

			for (di = 0, si = 2; si < tbuf[0]; si += 2) {
        		if ((tbuf[si] & 0x80) || (tbuf[si + 1])) /* non-ASCII */
		    	    string_buffer[di++] = '?';
		        else
			        string_buffer[di++] = tbuf[si];
	        }
        	string_buffer[di] = 0;
			printf("String descriptor %u: \"%s\"\n", a[i], string_buffer);

			
		} else {
			fprintf(stderr, "libusb_get_string_descriptor wValue=%i error %d\n", a[i], r);
 		    goto out;
		}
		
		// after string_descriptor 0303 comes SET_CONFIGURATION = 03
		if (a[i]==3){
		 	r = libusb_set_configuration(devh, 3);
        	if (r < 0) {
 		        fprintf(stderr, "libusb_set_configuration error %d\n", r);
 		        goto out;
 	        }
 	        printf("Successfully set usb configuration 3\n");
		
		}
	}
	
	
 	//libusb_reset_device(devh);
 
 	// Claiming of interfaces is a purely logical operation; 
    // it does not cause any requests to be sent over the bus. 
 	r = libusb_claim_interface(devh, 0);
 	if (r <0) {
 		fprintf(stderr, "libusb_claim_interface 0 error %d\n", r);
 		goto out;
 	}
 	
 	r = libusb_claim_interface(devh, 1);
 	if (r < 0) {
 		fprintf(stderr, "libusb_claim_interface 1 error %d\n", r);
 		goto out;
 	}

 	r = libusb_claim_interface(devh, 2);
 	if (r < 0) {
 		fprintf(stderr, "libusb_claim_interface 2 error %d\n", r);
 		goto out;
 	}



 	printf("Successfully claimed interface 0,1,2\n");
	
  	
/* Flir config
01 0b 01 00 01 00 00 00 c4 d5
0 bmRequestType = 01
1 bRequest = 0b
2 wValue 0001 type (H) index (L)    stop=0/start=1
4 wIndex 01                         interface 1/2
5 wLength 00
6 Data 00 00

libusb_control_transfer (*dev_handle, bmRequestType, bRequest, wValue,  wIndex, *data, wLength, timeout)
*/
 	
	printf("stop interface 2 FRAME\n");
 	unsigned char data[2]={0,0};
 	r = libusb_control_transfer(devh,1,0x0b,0,2,data,0,100);
 	if (r < 0) {
 		fprintf(stderr, "Control Out error %d\n", r);
 		return r;
 	}

	printf("stop interface 1 FILEIO\n");
 	r = libusb_control_transfer(devh,1,0x0b,0,1,data,0,100);
 	if (r < 0) {
 		fprintf(stderr, "Control Out error %d\n", r);
 		return r;
 	}
 	
	unsigned char buf[1048576]; 
    int actual_length;

 	int state = 1;
 	int count = 0;
 	time_t now;
 	char EP81_error[50]="", EP83_error[50]="",EP85_error[50]=""; 
    while (1)
    {
    	

//	printf("count= %d state = %d\n ",count,state);
/* 	r = libusb_claim_interface(devh, 1);
 	if (r < 0) {
 		fprintf(stderr, "libusb_claim_interface 1  error %d\n", r);
 		goto out;
 	}
*/
        switch(state) {
        
         case 1:
//          SETUP: 76.0  
//          DATA0: 01 0b 01 00 01 00 00 00 c4 d5         
//          0 bmRequestType = 01
//          1 bRequest = 0b                     SET_INTERFACE
//          2 wValue 0001 type (H) index (L)           Alternate Setting 1
//          4 wIndex 01                                    Interface 1
//          5 wLength 00 
         	
         	printf("\nstart interface 1 FILEIO\n");
         	unsigned char data[2]={0,0};
         	r = libusb_control_transfer(devh,1,0x0b,1,1,data,0,100);
 	        if (r < 0) {
 		        fprintf(stderr, "Control Out error %d\n", r);
 		        return r;
 	        }
 	        now = time(0); // Get the system time
            printf("\n:xx %s",ctime(&now));
 	        state = 2;
            break;
        
        
        case 2:
            count++;
            if (count>10)
            {
                count=0;
                state = 3;
            } 
            break;
        
        
        case 3:
         	printf("\nask for CameraFiles.zip on EP 0x83:\n");     
         	now = time(0); // Get the system time
            printf("\n: %s",ctime(&now));
   
            int transferred = 0;
            char my_string[128];

            int length = 16;
            unsigned char my_string2[16]={0xcc,0x01,0x00,0x00,0x01,0x00,0x00,0x00,0x41,0x00,0x00,0x00,0xF8,0xB3,0xF7,0x00};
            printf("\nEP 0x02 to be sent Hexcode: %i Bytes[",length);
            for (int i = 0; i < length; i++) {
                printf(" %02x", my_string2[i]);

            }
            printf(" ]\n");
    
            r = libusb_bulk_transfer(devh, 2, my_string2, length, &transferred, 0);
            if(r == 0 && transferred == length)
            {
                printf("\nWrite successful!");
            }
            else
                printf("\nError in write! res = %d and transferred = %d\n", r, transferred);


            //strcpy(  my_string, "{\"type\":\"setOption\",\"data\":{\"option\":\"autoFFC\",\"value\":true}}");
            //strcpy(  my_string,"{\"type\":\"readFile\",\"data\":{\"streamIdentifier\":10}}");
    
            strcpy(  my_string,"{\"type\":\"openFile\",\"data\":{\"mode\":\"r\",\"path\":\"CameraFiles.zip\"}}");
    
            length = strlen(my_string)+1;
            printf("\nEP 0x02 to be sent: %s", my_string);
    
            // avoid error: invalid conversion from ‘char*’ to ‘unsigned char*’ [-fpermissive]
            unsigned char *my_string1 = (unsigned char*)my_string;
            //my_string1 = (unsigned char*)my_string;
            
            r = libusb_bulk_transfer(devh, 2, my_string1, length, &transferred, 0);
            if(r == 0 && transferred == length)
            {
                printf("\nWrite successful!");
                printf("\nSent %d bytes with string: %s\n", transferred, my_string);
            }
            else
                printf("\nError in write! res = %d and transferred = %d\n", r, transferred);
 
 //---------
            length = 16;
            unsigned char my_string3[16]={0xcc,0x01,0x00,0x00,0x01,0x00,0x00,0x00,0x33,0x00,0x00,0x00,0xef,0xdb,0xc1,0xc1};
            printf("\nEP 0x02 to be sent Hexcode: %i Bytes[",length);
            for (int i = 0; i < length; i++) {
                printf(" %02x", my_string3[i]);

            }
            printf(" ]\n");
    
            r = libusb_bulk_transfer(devh, 2, my_string3, length, &transferred, 0);
            if(r == 0 && transferred == length)
            {
                printf("\nWrite successful!");
            }
            else
                printf("\nError in write! res = %d and transferred = %d\n", r, transferred);


            //strcpy(  my_string, "{\"type\":\"setOption\",\"data\":{\"option\":\"autoFFC\",\"value\":true}}");
            strcpy(  my_string,"{\"type\":\"readFile\",\"data\":{\"streamIdentifier\":10}}");
            //strcpy(  my_string,"{\"type\":\"openFile\",\"data\":{\"mode\":\"r\",\"path\":\"CameraFiles.zip\"}}");
            length = strlen(my_string)+1;
            printf("\nEP 0x02 to be sent %i Bytes: %s", length, my_string);
    
            // avoid error: invalid conversion from ‘char*’ to ‘unsigned char*’ [-fpermissive]
            my_string1 = (unsigned char*)my_string;
            //my_string1 = (unsigned char*)my_string;
            
            r = libusb_bulk_transfer(devh, 2, my_string1, length, &transferred, 0);
            if(r == 0 && transferred == length)
            {
                printf("\nWrite successful!");
                printf("\nSent %d bytes with string: %s\n", transferred, my_string);
            }
            else
                printf("\nError in write! res = %d and transferred = %d\n", r, transferred);
 
 
            // go to next state
            now = time(0); // Get the system time
            printf("\n: %s",ctime(&now));

            count = 0;
            state = 4;            
            break;
            

        case 4:

            count++;
            if (count>20)
            {
                now = time(0); // Get the system time
                printf("\n: %s",ctime(&now));

                count=0;
                state = 5;      // change to next step is 5 !!!!!!!!!!!

            } 
            break;



        case 5:
         	printf("\nAsk for video stream, start EP 0x85:\n");        

            r = libusb_control_transfer(devh,1,0x0b,1,2,data, 2,200);
            if (r < 0) {
                fprintf(stderr, "Control Out error %d\n", r);
                return r;
            };
            state = 6;
            break;

        case 6:
            // endless loop
            r = libusb_bulk_transfer(devh, 0x85, buf, sizeof(buf), &actual_length, 100); 
//            print_bulk_result("0x85",EP85_error, r, actual_length, buf);
	    printf("read from 85\n");
            break;

        }    

        // poll Endpoints 0x81, 0x83, 0x85
        
	    
        r = libusb_bulk_transfer(devh, 0x81, buf, sizeof(buf), &actual_length, 100); 
        print_bulk_result("0x81",EP81_error, r, actual_length, buf);

        r = libusb_bulk_transfer(devh, 0x83, buf, sizeof(buf), &actual_length, 100); 
        print_bulk_result("0x83",EP83_error, r, actual_length, buf);
                 
    }

 	libusb_release_interface(devh, 0);
 out:
 //close the device
 	libusb_reset_device(devh);
 	libusb_close(devh);
 	libusb_exit(NULL);
 	return r >= 0 ? r : -r;
 }
